/* Last edited on 2004-09-26 20:40:05 by stolfi */

#define PROG_NAME "espectro"
#define PROG_DESC "Fourier analysis of DNA sequences."
#define PROG_HELP "NSEQ NCUT SIGMA0 SHIFT SCALE < PAIRag.txt > PAIRep.txt"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <fftw3.h>

/*
  Entrada (stdin): 
    Arquivo contendo duas sequencias alinhadas, p.ex. "{PAIR}_1_2ag.txt"
  
  Saida (stdout):
    Arquivo contendo os espectros de potncia de cada seq individualmente
    e dos sinais mdio e diferena, p.ex. "s{SCALE}-{NCUT}-{PAIR}_1_2ep.txt".
    
    Linhas iniciais: comentrios iniciados com "#"
    
    Linhas do meio: {floor(nSmp/2)+1} linhas com espectro de potncia
    
      {k} {\tilde A_k} {\tilde B_k} {\tilde M_k} {\tilde D_k}
    
    As linhas finais contm os valores de 
    
      {\Re A_0} {#A - #T} {\Im A_0} {#C - #G} 
      {\Re B_0} {#A - #T} {\Im B_0} {#C - #G}   
      {\Re M_0} {#A - #T} {\Im M_0} {#C - #G}   
      {\Re D_0} {#A - #T} {\Im D_0} {#C - #G}   
*/

#define PI 3.141592653589793238

int main (int argc, char **argv);

void le_cadeia(
    int nSeq, 
    int nCut, 
    double sigma, 
    int shift, 
    int step, 
    int nSmp,
    fftw_complex *smp
  );
  /* 
    L cadeia {r} com {nSeq} bases de {stdin}, pega as {nCut} bases
    centrais, converte para complexos, filtra com parmetro {sigma},
    sub-amostra {smp[i] = a[shift + i*step]}. */

void processa_cadeia(
    int nSmp, 
    fftw_complex *smp,
    double *pot,
    double *Re0,
    double *Im0
  );
  /* Calcula DFT da cadeia {smp[0..nSmp-1][0..1]}, armazena espectro
    de potncia em {pot[0..FMAX]} onde {FMAX = floor(nSmp/2)}. 
    Retorna tambm as partes real {*Re0} e imaginria {*Im0} 
    do coeficiente da DFT de freq. zero. */

void calcula_pot(int nSmp, fftw_complex *dft, double *pot);
  /* Calcula o espectro de potncia {pot[0..FMAX]} a partir
    da DFT {dft[0..nSmp-1][0..1]} */

double gauss(double sigma, double dev);
  /* Gaussiana no normalizada: */
double gauss_sum(double sigma);
  /* Calcula fator de normalizao do filtro gaussiano: */
double gauss_dev(double sigma);
  /* Calcula desvio padro do filtro gaussiano: */

double fltround(double x);
  /* Arredonda para int mais proximo */

/*********************************************************************/

void le_cadeia(
    int nSeq, 
    int nCut, 
    double sigma, 
    int shift, 
    int step, 
    int nSmp,
    fftw_complex *smp
  )
  {
    int nr; /* Conta bases lidas. */
    int discard = (nSeq - nCut)/2;  /* Num bases iniciais a descartar. */
    int k;

    if(nCut != nSmp * step)
      { fprintf(stderr, "Bad nCut = %d nSmp = %d  step = %d\n", nCut, nSmp, step); }
    if((shift < 0) || (shift >= step))
      { fprintf(stderr, "Bad shift = %d  step = %d\n", shift, step); }

    fftw_complex *cut  = fftw_malloc(sizeof(fftw_complex)*nCut);

    nr = 0;
    while (nr < nSeq)
      {
        char base = getc(stdin);
        if ((base != '\n') && (base != ' '))
          {
            /* Mapeia bases {A,T,C,G} para n?meros complexos {?1,?I}: */
            double re, im;
            switch(base)
              {
                case 'A': re = +1; im = 00; break;
                case 'T': re = -1; im = 00; break;
                case 'C': re = 00; im = +1; break;
                case 'G': re = 00; im = -1; break;
                case '-': re = 00; im = 00; break;
                default:
                  fprintf(stderr, "Invalid base (%c)\n", base);
                  exit(1);
              }
            if ((nr >= discard) && (nr-discard < nCut))
              { int ic = nr - discard;
                cut[ic][0] = re;
                cut[ic][1] = im;
                /* fprintf(stdout, "# %6d  %+4.1f %+4.1f\n", ic, cut[ic][0], cut[ic][1]); */
              }
            nr++;
          }
      }

    /* Parmetros para filtragem: */
    double gnorm, gdev;
    if (sigma == 0)
      { gnorm = 1.0; gdev = 0.0; }
    else
      { 
        fprintf(stdout, "# sigma = %8.5f\n", sigma);
        gnorm = gauss_sum(sigma);
        fprintf(stdout, "# gnorm = %8.5f\n", gnorm);
        gdev = gauss_dev(sigma);
        fprintf(stdout, "# gdev = %8.5f\n", gdev);
      }

     /* Unidade de quantizao: */
    double unit = 0.0001; 
    fprintf(stdout, "# unit = %22.15e\n", unit);
    /* Filtra cadeia {cut} com gaussiana, amostra, coloca em {smp} */
    int is;
    for(is = 0; is < nSmp; is++)
      {
        int ic = (shift + step*is) % nCut; 
        while (ic < 0) { ic += nCut; }
        for (k = 0; k < 2; k++)
          { double val;
            if (sigma == 0)
              { val = cut[ic][k]; /* Filtragem trivial */ }
            else
              { 
                int j = 0;
                double wj = gauss(sigma, (double)j)/gnorm;
                double sum = wj * cut[ic][k];
                j ++;
                while ((wj = gauss(sigma, (double)j)/gnorm) > 1.0e-12)
                  { int ijc = (ic - j) % nCut; 
                    while (ijc < 0) { ijc += nCut; }
                    sum += wj * cut[ijc][k];
                    ijc = (ic + j) % nCut;
                    sum += wj * cut[ijc][k];
                    j++;
                  }
                val = sum;
              }
            /* Quantize to five decimals: */
            smp[is][k] = fltround(val/unit)*unit;
          }
        /* fprintf(stdout, "# %6d  %+10.7f %+10.7f\n", is, smp[is][0], smp[is][1]);  */
      }
    free(cut);
  }

void processa_cadeia(
    int nSmp, 
    fftw_complex *smp,
    double *pot,
    double *Re0,
    double *Im0
  )
  {
    double norm = 1.0/nSmp;
    /* Internal working vectors: */
    fftw_complex *fft_in = fftw_malloc(sizeof(fftw_complex)*nSmp);
    fftw_complex *fft_ot = fftw_malloc(sizeof(fftw_complex)*nSmp);
    fftw_plan plan = fftw_plan_dft_1d(nSmp, fft_in, fft_ot, FFTW_FORWARD, FFTW_ESTIMATE);
    int i;
    for(i=0; i < nSmp; i++) { fft_in[i][0]=smp[i][0]; fft_in[i][1]=smp[i][1]; }
    fftw_execute(plan);
    *Re0 = sqrt(norm)*fft_ot[0][0]; *Im0 = sqrt(norm)*fft_ot[0][1];
    calcula_pot(nSmp, fft_ot, pot);
    free(fft_in); free(fft_ot);
  }

void calcula_pot(int nSmp, fftw_complex *dft, double *pot)
  {
    int frMax = (int)(nSmp/2);    /* Frequencia maxima. */
    double norm = 1.0/nSmp;
    int i;
    double pot_pos, pot_neg;

    /* Frequencia {0} */
    i = 0;
    pot_pos = dft[i][0]*dft[i][0] + dft[i][1]*dft[i][1];
    pot[i] = norm * pot_pos;

    /* Frequencias {k} com {0 < k < nSmp/2}: */
    for (i = 1; 2*i < nSmp; i++)
      {
        int j = nSmp-i;
        pot_pos = dft[i][0]*dft[i][0] + dft[i][1]*dft[i][1];
        pot_neg = dft[j][0]*dft[j][0] + dft[j][1]*dft[j][1];
        pot[i] = norm * (pot_pos + pot_neg);
      }

    /* Frequencia {N/2}: */
    if ((nSmp%2) == 0) //se nSmp for impar o frMax foi somado ao frMax-1
      {
        i = frMax;
        pot_pos = dft[i][0]*dft[i][0] + dft[i][1]*dft[i][1];
        pot[i] = norm * pot_pos;
      }
  }

double gauss(double sigma, double dev)
  {
    double mag = 1/(sigma*sqrt(2*PI));
    return mag * exp(-dev*dev/(2*sigma*sigma));
  }

double gauss_sum(double sigma)
  {
    int j = 0;
    double wj = gauss(sigma, (double)j);
    double sum = wj;
    j ++;
    while ((wj = gauss(sigma, (double)j)) > 1.0e-15)
      { sum += 2*wj;
        j++;
      }
    return sum;
  }

double gauss_dev(double sigma)
  {
    int j = 0;
    double wj = gauss(sigma, (double)j);
    double sum = 0;
    j ++;
    while ((wj = gauss(sigma, (double)j)) > 1.0e-15)
      { sum += 2*wj*(double)j*(double)j;
        j++;
      }
    return sqrt(sum/gauss_sum(sigma));
  }

double fltround(double x)
  {
    double rax = floor(fabs(x) + 0.5);
    return (x < 0.0 ? -rax : rax);
  }

int main (int argc, char **argv)
{
  /* Anlise dos parmetros de chamada */
  if (argc != 6)
    {
      fprintf(stderr, "Parametros errados!\n");
      fprintf(stderr, "%s %s\n", PROG_NAME, PROG_HELP);
      return(1);
    }
  int nSeq = atoi(argv[1]); /* Num bases nas cadeidas de entrada. */
  int nCut = atoi(argv[2]); /* Num bases a usar. */
  double sigma_zero = atof(argv[3]);  /* Amaciamento para escala 0 -> 1. */
  int shift = atoi(argv[4]); /* Deslocamento em bases na sub-amostragem. */
  int scale = atoi(argv[5]); /* Escala desejada. */

  /* Parmetros derivados: */
  int step = (1 << scale);    /* Passo de sub-amostragem. */
  fprintf(stdout, "# step = %d\n", step);

  int nSmp = nCut/step;       /* Numero de amostras nas cadeias reduzidas. */
  fprintf(stdout, "# nSmp = %d\n", nSmp);

  int frMax = (int)(nSmp/2);  /* Freqncia mxima. */
  int nFreqs = frMax + 1;
  fprintf(stdout, "# frMax = %d\n", frMax);
  fprintf(stdout, "# nFreqs = %d\n", nFreqs);

  /* Reduce {shift} to {-floor(step/2) .. +floor(step/2)} modulo {step}: */
  /* If {step} is even, give equal chances to the extremes: */
  shift = shift % (2*step);
  while (shift < 0) { shift += 2*step; }
  if (shift > step) { shift -= step; }
  if (2*shift > step) { shift -= step; }
  if (2*shift < -step) { shift += step; }
  fprintf(stdout, "# shift = %d\n", shift);
  int shift_a = (shift < 0 ? -shift : 0);
  int shift_b = (shift > 0 ?  shift : 0);
  fprintf(stdout, "# shift_a = %d\n", shift_a);
  fprintf(stdout, "# shift_b = %d\n", shift_b);

  /* Fator de normalizao para o espectro de potncia: */
  double norm = 1.0/nSmp;
  
  /* Parametro do ncleo da gaussiana */
  double sigma = (scale == 0 ? 0.0 : sigma_zero * (double)(1<<scale)); 
  fprintf(stdout, "# sigma = %12.7f\n", sigma);

  /* Sinais filtrados e sub-amostrados: */
  fftw_complex *smp_a = fftw_malloc(sizeof(fftw_complex)*nSmp); /* Sinal {a} */
  fftw_complex *smp_b = fftw_malloc(sizeof(fftw_complex)*nSmp); /* Sinal {b} */
  fftw_complex *smp_d = fftw_malloc(sizeof(fftw_complex)*nSmp); /* Sinal diferena {d} */
  fftw_complex *smp_m = fftw_malloc(sizeof(fftw_complex)*nSmp); /* Sinal mdio {m} */

  /* Espectros de potncia: */
  double *pot_a = (double*) malloc(sizeof(double)*nFreqs);   /* Espectro de {a} */
  double *pot_b = (double*) malloc(sizeof(double)*nFreqs);   /* Espectro de {b} */
  double *pot_d = (double*) malloc(sizeof(double)*nFreqs);   /* Espectro de {d} */
  double *pot_m = (double*) malloc(sizeof(double)*nFreqs);   /* Espectro de {m} */

  double Re0_a, Im0_a, Re0_b, Im0_b;  /* Componentes constantes de m e d */
  le_cadeia(nSeq, nCut, sigma, shift_a, step, nSmp, smp_a);
  processa_cadeia(nSmp, smp_a, pot_a, &Re0_a, &Im0_a);

  getc(stdin); //consome '\n'

  le_cadeia(nSeq, nCut, sigma, shift_b, step, nSmp, smp_b);
  processa_cadeia(nSmp, smp_b, pot_b, &Re0_b, &Im0_b);

  /* Calcula a mdia e diferenca entre S e T */
  double Re0_m, Im0_m, Re0_d, Im0_d;  /* Componentes constantes de m e d */
  int i;
  for(i = 0; i < nSmp; i++)
  {
    smp_m[i][0]=(smp_a[i][0]+ smp_b[i][0])/2;
    smp_m[i][1]=(smp_a[i][1]+ smp_b[i][1])/2;

    smp_d[i][0]=smp_a[i][0] - smp_b[i][0];
    smp_d[i][1]=smp_a[i][1] - smp_b[i][1];
  }
  
  processa_cadeia(nSmp, smp_d, pot_d, &Re0_d, &Im0_d);
  processa_cadeia(nSmp, smp_m, pot_m, &Re0_m, &Im0_m);
  
  /* Impresso */
  int k;
  fprintf (stdout, "# k \\tilde A_k \\tilde B_k \\tilde M_k \\tilde D_k\n");
  fprintf (stdout, "\n");
  for (k = 0; k <= frMax; k++)
    {
      fprintf(stdout, "%6d %16.8f %16.8f %16.8f %16.8f\n",
        k, pot_a[k], pot_b[k], pot_m[k], pot_d[k]
      );
    }
  fprintf (stdout, "\n");
  fprintf (stdout, "# \\Re(X_0)  #A-#T  \\Im(X_0)  #C-#G\n");
  fprintf (stdout, "\n");
  fprintf (stdout, " %12.8f %12.2f  %12.8f %12.2f # a\n", 
    Re0_a, Re0_a/sqrt(norm), Im0_a, Im0_a/sqrt(norm)
  );
  fprintf (stdout, " %12.8f %12.2f  %12.8f %12.2f # b\n", 
    Re0_b, Re0_b/sqrt(norm), Im0_b, Im0_b/sqrt(norm)
  );
  fprintf (stdout, " %12.8f %12.2f  %12.8f %12.2f # m\n", 
    Re0_m, Re0_m/sqrt(norm), Im0_m, Im0_m/sqrt(norm)
  );
  fprintf (stdout, " %12.8f %12.2f  %12.8f %12.2f # d\n", 
    Re0_d, Re0_d/sqrt(norm), Im0_d, Im0_d/sqrt(norm)
  );

  fflush(stdout);

  return 0;
}
