/* Last edited on 2004-09-26 20:32:20 by stolfi */ 

#define PROG_NAME "qtinfo"
#define PROG_HELP "DIR NBASES SCALE < PAIRNAMES.lst > OUT.txt"

#define _GNU_SOURCE
#include <stdio.h>
#include <math.h>
#include <fftw3.h>
#include <string.h>
#include <stdlib.h>

/*
   Parmetros: 
     DIR - diretrio onde esto dados de entrada
     NBASES - nmero de bases nas cadeias de nvel 0
     SCALE - escala de amostragem (0 = original)
   
   Entrada principal (stdin): arquivo com nomes de pares.
     
   Um arquivo de entrada "{DIR}/s{SCALE}-{PAIR}.txt" para cada 
   par, no formato:
   
     Linhas que comeam com "#" so ignoradas.
     Deve conter FRMAX+1 linhas sem "#" contendo cada uma
       {k} {\tilde A_k} {\tilde B_k} {\tilde M_k} {\tilde D_k} 
     para {k = 0..FRMAX}, onde {FRMAX = floor(NBASES/2^SCALE/2)}.
     Seguidas de 4 linhas com as componentes constantes dos quatro
     sinais associados ao par:
      {\Re A_0} {#A - #T} {\Im A_0} {#C - #G} 
      {\Re B_0} {#A - #T} {\Im B_0} {#C - #G}   
      {\Re M_0} {#A - #T} {\Im M_0} {#C - #G}   
      {\Re D_0} {#A - #T} {\Im D_0} {#C - #G}   
     
   Sada: Arquivo contendo:
   
     para cada frequencia {k}
       a mdia {\tilde M_k} dos espectros de potncia dos sinais mdios,
       a varincia estimada de {\Op M_k}
       a mdia{\tilde D_k}  dos espectros de potncia dos sinais diferena,
       a varincia estimada de {\Op D_k}
       a informao mtua por varivel {I(\Op A_k -> \Op B_k)}. 
       a quantidade mdia de informao da freqncia absoluta.
     
     No final:
       Quantidade total de informao mtua.
       Quantidade de informao mtua por base.
       Seguido da tabela formatada em LaTeX.
*/

void pula_comentarios(FILE *arq);
/* Pula brancos e comentrios. */

void pula_comentarios(FILE *arq)
{
  int ch;
  do { 
    do { ch = fgetc(arq); } while ((ch == ' ') || (ch == '\011'));
    if (ch == '#')
      { do { ch = fgetc(arq); } while ((ch != '\n') && (ch != EOF)); }
  } while (ch == '\n');
  if (ch != EOF) { ungetc(ch, arq); }
}

int main (int argc, char **argv)
{

  char  nome[21];
  int 	k,pares;
  double total_info;
  double sum_Re0_D = 0, sum_Im0_D = 0, sum_Re0_M = 0, sum_Im0_M = 0;

  if (argc != 4)
    { printf("%s: parametros errados!\n%s %s\n", PROG_NAME, PROG_NAME, PROG_HELP);
      exit(1);
    }
  char *dir = argv[1];
  int nBases = atoi(argv[2]);
  int scale = atoi(argv[3]);
  
  int nSmp = nBases >> scale;
  if (nBases != (nSmp << scale)) 
    { printf("%s: bad nBases = %d scale = %d\n", PROG_NAME, nBases, scale);
      exit(1);
    }
  double fnorm = sqrt(1.0/nSmp); /* Fator de normalizao da FFT. */

  int frMax = nSmp/2;
  int nFreqs = frMax + 1;

  double sum_tilde_M[nFreqs], sum_tilde_D[nFreqs];

  for(k=0; k<=frMax; k++)
    { sum_tilde_M[k] = sum_tilde_D[k] = 0.0; }

  pares = 0;

  while(fscanf(stdin,"%20s", nome) > 0)
  {
    pares++;
    char *nomeArq = NULL;
    asprintf(&nomeArq, "%s/s%d-%d-%s.txt", dir, scale, nBases, nome);
    FILE *arq = fopen(nomeArq,"rb");
    fprintf(stdout, "# file = %s\n", nomeArq); 
    if (!arq)
      { fprintf(stderr, "\nErro na abertura do arquivo %s\n", nomeArq); exit(1); }
      
    for(k=0; k<=frMax; k++)
      { 
        double A, B, D, M;
        int kk;
        pula_comentarios(arq);
        if (fscanf(arq,"%d %lf %lf %lf %lf\n", &kk, &A, &B, &M, &D) != 5)
          { fprintf(stderr, "\nErro de formato no arquivo %s k = %d\n", nomeArq, k);
            exit(1);
          } 
        if (kk != k)
          { fprintf(stderr, "\nErro de seqncia no arquivo %s k = %d\n", nomeArq, k);
            exit(1);
          }
        /* fprintf(stdout, "# %6d %16.8f %16.8f \n", k, M, D); */ 
        sum_tilde_M[k] = sum_tilde_M[k] + M;
        sum_tilde_D[k] = sum_tilde_D[k] + D;
      }
    /* Dados de termos constantes: */
    pula_comentarios(arq);
    double Re0_A, AT_A, Im0_A, CG_A;
    fscanf(arq,"%lf %lf %lf %lf\n", &Re0_A, &AT_A, &Im0_A, &CG_A);
    
    pula_comentarios(arq);
    double Re0_B, AT_B, Im0_B, CG_B;
    fscanf(arq,"%lf %lf %lf %lf\n", &Re0_B, &AT_B, &Im0_B, &CG_B);
    
    pula_comentarios(arq);
    double Re0_M, AT_M, Im0_M, CG_M;
    fscanf(arq,"%lf %lf %lf %lf\n", &Re0_M, &AT_M, &Im0_M, &CG_M);
    
    pula_comentarios(arq);
    double Re0_D, AT_D, Im0_D, CG_D;
    fscanf(arq,"%lf %lf %lf %lf\n", &Re0_D, &AT_D, &Im0_D, &CG_D);
    
    sum_Re0_M += Re0_M;
    sum_Im0_M += Im0_M;
    sum_Re0_D += Re0_D;
    sum_Im0_D += Im0_D;
    fclose(arq);
    free(nomeArq);
  }
  
  /* Calculando espectro de potncia mdio e : */
  double avg_tilde_M[nFreqs], avg_tilde_D[nFreqs];
  for(k=0;k<=frMax;k++)
  {
    avg_tilde_M[k] = sum_tilde_M[k] / pares;
    avg_tilde_D[k] = sum_tilde_D[k] / pares;
  }

  /* Calculando mdias dos coefs indeps: */
  double avg_Re0_M = sum_Re0_M / pares;
  double avg_Im0_M = sum_Im0_M / pares;
  double avg_Re0_D = sum_Re0_D / pares;
  double avg_Im0_D = sum_Im0_D / pares;

  /* Estimando varincias de {\Op M_k} {\Op D_k} */
  double VM[nFreqs], VD[nFreqs];
  VM[0] = 0.5 * (sum_tilde_M[0] - (avg_Re0_M*avg_Re0_M + avg_Im0_M*avg_Im0_M)*pares)/(pares-1);
  VD[0] = 0.5 * (sum_tilde_D[0] - (avg_Re0_D*avg_Re0_D + avg_Im0_D*avg_Im0_D)*pares)/(pares-1);
  for(k = 1; k <= frMax; k++)
  {
    /* Varincias: */
    if (2*k == nSmp)
      { VM[k] = 0.5 * avg_tilde_M[k];
        VD[k] = 0.5 * avg_tilde_D[k];
      }
    else
      { VM[k] = 0.25 * avg_tilde_M[k];
        VD[k] = 0.25 * avg_tilde_D[k];
      }
  }
  
  /* Calculando informao mtua: */
  total_info=0;
  double I[nFreqs], IAB[nFreqs];
  for(k=0;k<=frMax;k++)
  { 
    /* Informao mtua {I(\Op A_k \to \Op B_k)}: */
    double x = (VM[k]/VD[k]) + 0.5 + 0.0625*(VD[k]/VM[k]);
    IAB[k] = 0.5 * log(x)/log(2);
    
    /* Informao mtua por freqncia: */
    if ((k == 0) || (2*k == nSmp))
      { I[k] = 2*IAB[k]; }
    else
      { I[k] = 4*IAB[k]; }
      
    /* Acumula informao total: */
    total_info = total_info + I[k];
  }

  /* Impresso dos dados: */
  fprintf(stdout, "#  k \\tilde M_k \\hat M_k  \\tilde D_k \\hat D_k  I(Op A_k -> Op B_k) I_k\n");
  fprintf(stdout, "\n");
  for(k=0; k<=frMax; k++)
    {
      fprintf(stdout, 
        "%6d %8.5f %8.5f   %8.5f %8.5f   %8.5f %8.5f \n", 
        k, avg_tilde_M[k], VM[k], avg_tilde_D[k], VD[k], IAB[k], I[k]
      );
    }
  fprintf(stdout, "\n");
  fprintf(stdout, "# Escala = %d\n", scale);
  fprintf(stdout, "# Informao total mtua = %8.5f\n", total_info);
  fprintf(stdout, "# Informao mtua por elemento = %8.5f\n", total_info/nSmp);
  fprintf(stdout, "# Informao mtua por base = %8.5f\n", total_info/nBases);
  
  fprintf(stdout, "\n");
  fprintf(stdout, "# \\Re(X_0) #A-#T \\Im(X_0) #C-#G\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "#  %12.8f %12.2f  %12.8f %12.2f # m\n", 
    avg_Re0_M, avg_Re0_M/fnorm, avg_Im0_M, avg_Im0_M/fnorm
  );
  fprintf(stdout, "#  %12.8f %12.2f  %12.8f %12.2f # d\n", 
    avg_Re0_D, avg_Re0_D/fnorm, avg_Im0_D, avg_Im0_D/fnorm
  );
  
  fprintf(stdout, "\n");
  fprintf(stdout, "#  k \\hat M_k  \\hat D_k  I_k\n");
  fprintf(stdout, "\n");

  /* Impresso dos dados em LaTeX: */
  int frCodon = nSmp/3;
  int printed_dots = 1;
  for(k=0; k<=frMax; k++)
    {
      if ((k <= 5) || ((k >= frCodon-3) && (k <= frCodon+3)) || (k >= frMax-5))
        { 
          fprintf(stdout, 
            "## %6d & %8.3f & %8.3f & %8.3f \\\\ \n", 
            k, VM[k], VD[k], I[k]
          );
          printed_dots = 0;
        }
      else if (! printed_dots)
        { 
          fprintf(stdout, 
            "## %6s & %8s & %8s & %8s \\\\ \n", 
               "\\ldots",  "\\ldots",  "\\ldots",  "\\ldots" 
          );
          printed_dots = 1;
        }
    }

  fflush(stdout);

  return 0;

}
